/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/semaphore.c
 * @brief functions related to threading and synchronization
 *
 * In particular, functions for mutexes, semaphores 
 * and thread creation are provided.
 **/

#include "config.h"

#include "util/semaphore.h"
#include "util/logging.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"

#include <errno.h>
#include <string.h>
#ifdef SOMEBSD
# include <pthread_np.h>
#endif

#define DEBUG_SEMUPDOWN NO

/**
 * Shall we use error-checking (slow) 
 * mutexes (e.g. for debugging)
 **/
#define USE_CHECKING_MUTEX 1


#ifndef PTHREAD_MUTEX_NORMAL
#ifdef PTHREAD_MUTEX_TIMED_NP
#define PTHREAD_MUTEX_NORMAL PTHREAD_MUTEX_TIMED_NP
#else
#define PTHREAD_MUTEX_NORMAL NULL
#endif
#endif

/**
 * This prototype is somehow missing in various Linux pthread
 * include files. But we need it and it seems to be available
 * on all pthread-systems so far. Odd.
 **/
extern int pthread_mutexattr_setkind_np(pthread_mutexattr_t *attr, int kind);

/* ********************* public methods ******************* */

void create_mutex_(Mutex * mutex) {
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
#if USE_CHECKING_MUTEX
#ifdef LINUX
  pthread_mutexattr_setkind_np(&attr,
			       PTHREAD_MUTEX_ERRORCHECK_NP);  
#endif
#else
  pthread_mutexattr_setkind_np(&attr,
			       (int)PTHREAD_MUTEX_NORMAL);
#endif
  if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0)
    errexit("FATAL: creation of mutex failed (%s)",
	    strerror(errno));
}

void create_recursive_mutex_(Mutex * mutex) {
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
#ifdef LINUX
  pthread_mutexattr_setkind_np(&attr,  
			       PTHREAD_MUTEX_RECURSIVE_NP);
#elif SOMEBSD
  pthread_mutexattr_setkind_np(&attr,
			       PTHREAD_MUTEX_RECURSIVE);
#elif SOLARIS
  pthread_mutexattr_settype(&attr,
			    PTHREAD_MUTEX_RECURSIVE);
#elif CYGWIN
  pthread_mutexattr_settype(&attr,
			    PTHREAD_MUTEX_RECURSIVE);
#endif
  if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0)
    errexit("FATAL: creation of recursive mutex failed");
}

void destroy_mutex_(Mutex * mutex) {
  pthread_mutex_destroy(&mutex->pthreadMutex);
}

void mutex_lock_(Mutex * mutex,
		 const char * filename,
		 const int line) {
  int return_value;
  
  if ((return_value = pthread_mutex_lock (&mutex->pthreadMutex)) != 0)
    errexit("FATAL: pthread_mutex_lock returned %d (%s) in %s:%d\n",
	    return_value, 
	    strerror(return_value),
	    filename,
	    line);  
}

void mutex_unlock_(Mutex * mutex,
		   const char * filename,
		   const int line) {
  int return_value;

  if ((return_value = pthread_mutex_unlock(&mutex->pthreadMutex)) != 0)
    errexit("FATAL: pthread_mutex_unlock returned %d (%s) in %s:%d\n",
	    return_value, 
	    strerror(return_value),
	    filename,
	    line);  
}

/**
 * function must be called prior to semaphore use -- handles
 * setup and initialization.  semaphore destroy (below) should
 * be called when the semaphore is no longer needed.
 **/
Semaphore * semaphore_new_(int value,
			   const char * filename,
			   const int linenumber) {
  Semaphore * s = xmalloc_(sizeof(Semaphore),
			   filename,
			   linenumber);
  s->v = value;
  MUTEX_CREATE(&(s->mutex));

  if (pthread_cond_init (&(s->cond), NULL) != 0)
    errexit("FATAL: pthread_cont_init failed in %s:%d\n",
	    filename,
	    linenumber);
  return s;
}

void semaphore_free_(Semaphore * s,
		     const char * filename,
		     const int linenumber) {
  MUTEX_DESTROY(&(s->mutex));
  pthread_cond_destroy(&(s->cond));
  xfree_(s,
	 filename,
	 linenumber);
}

/**
 * function increments the semaphore and signals any threads that
 * are blocked waiting a change in the semaphore.
 **/
int semaphore_up_(Semaphore * s,
		  const char * filename,
		  const int linenumber) {
  int value_after_op;
  int return_value;

  if (s == NULL)
    errexit("semaphore_up_ called with s == NULL in %s:%d\n",
	    filename, linenumber);
#if DEBUG_SEMUPDOWN
  LOG(LOG_DEBUG,
      "DEBUG: semaphore_up %d enter at %s:%d\n",
      (int)s,
      filename,
      linenumber);
#endif
  MUTEX_LOCK(&(s->mutex));  
  (s->v)++;
  value_after_op = s->v;  
  MUTEX_UNLOCK(&(s->mutex));
  if ((return_value = pthread_cond_signal(&s->cond) != 0))
      errexit("FATAL: pthread_cond_signal returned %d (%s) in %s:%d\n",
	      return_value, 
	      strerror(return_value),
	      filename,
	      linenumber);  
#if DEBUG_SEMUPDOWN
  LOG(LOG_DEBUG,
      "DEBUG: semaphore_up %d exit at %s:%d\n",
      (int)s,
      filename,
      linenumber);
#endif
  return (value_after_op);
}

/**
 * function decrements the semaphore and blocks if the semaphore is
 * <= 0 until another thread signals a change.
 **/
int semaphore_down_(Semaphore * s,
		    const char * filename,
		    const int linenumber) {
  int value_after_op;
  int return_value;
  
  if (s == NULL)
    errexit("semaphore_down_ called with s == NULL in %s:%d\n",
	    filename, linenumber);
#if DEBUG_SEMUPDOWN
  LOG(LOG_DEBUG,
      "DEBUG: semaphore_down %d enter at %s:%d\n",
      (int)s,
      filename,
      linenumber);
#endif
  MUTEX_LOCK(&(s->mutex));
  while (s->v <= 0) {
    if ((return_value = pthread_cond_wait(&s->cond, 
					  &s->mutex.pthreadMutex)) != 0)
      errexit("FATAL: pthread_cond_wait returned %d (%s) in %s:%d\n",
	      return_value,
	      strerror(return_value),
	      filename,
	      linenumber);  	    
  }
  (s->v)--;
  value_after_op = s->v;
  MUTEX_UNLOCK(&(s->mutex));  
#if DEBUG_SEMUPDOWN
  LOG(LOG_DEBUG,
      "DEBUG: semaphore_down %d exit at %s:%d\n",
      (int)s,
      filename,
      linenumber);
#endif
  return (value_after_op);
}

/**
 * Function decrements the semaphore. If the semaphore would become
 * negative, the decrement does not happen and the function returns
 * SYSERR. Otherwise OK is returned.
 **/
int semaphore_down_nonblocking_(Semaphore * s,
				const char * filename,
				const int linenumber) {
  if (s == NULL)
    errexit("semaphore_down_nonblocking called with s == NULL in %s:%d\n",
	    filename, linenumber);
  MUTEX_LOCK(&(s->mutex));
  if (s->v <= 0) {
    MUTEX_UNLOCK(&(s->mutex));
    return SYSERR;
  }
  (s->v)--;
  MUTEX_UNLOCK(&(s->mutex));  
  return OK;
}

/**
 * Create a thread. Use this method instead of pthread_create since
 * BSD may only give a 1k stack otherwise.
 *
 * @param handle handle to the pthread (for detaching, join)
 * @param main the main method of the thread
 * @param arg the argument to main
 * @param stackSize the size of the stack of the thread in bytes. 
 *        Note that if the stack overflows, some OSes (seen under BSD) 
 *        will just segfault and gdb will give a messed-up stacktrace.
 * @return see pthread_create 
 **/
int PTHREAD_CREATE(pthread_t * handle,
		   PThreadMain main,
		   void * arg,
		   size_t stackSize) {
  pthread_attr_t stack_size_custom_attr;

  pthread_attr_init(&stack_size_custom_attr);
  pthread_attr_setstacksize(&stack_size_custom_attr,
			    stackSize);
  return pthread_create(handle, 
			&stack_size_custom_attr,
			main, 
			arg);			    
}

void PTHREAD_JOIN(pthread_t handle,
		  void * ret) {
  int k;
  switch ((k=pthread_join(handle, ret))) {
  case 0: 
    return;
  case ESRCH:
    errexit("FATAL: pthread_join failed: ESRCH (%s)\n",
	    strerror(errno));
  case EINVAL:
    errexit("FATAL: pthread_join failed: EINVAL (%s)\n",
	    strerror(errno));
  case EDEADLK:
    errexit("FATAL: pthread_join failed: EDEADLK (%s)\n",
	    strerror(errno));
  default:
    errexit("FATAL: pthread_join failed: %d (%s)\n",
	    k,
	    strerror(errno));
    
  }
}

/* end of semaphore.c */
